"""Функции для получения информации о свечках."""

import aiohttp

from aiomoex import client, request_helpers
from aiomoex.request_helpers import (
    CANDLE_BORDERS,
    CANDLES,
    DEFAULT_BOARD,
    DEFAULT_ENGINE,
    DEFAULT_MARKET,
)


async def get_market_candle_borders(
    session: aiohttp.ClientSession,
    security: str,
    market: str = DEFAULT_MARKET,
    engine: str = DEFAULT_ENGINE,
) -> client.Table:
    """Получить таблицу интервалов доступных дат для всех режимов торгов.

    Описание запроса - https://iss.moex.com/iss/reference/156

    :param session:
        Сессия http соединения.
    :param security:
        Тикер ценной бумаги.
    :param market:
        Рынок - по умолчанию акции.
    :param engine:
        Движок - по умолчанию акции.

    :return:
        Список словарей, которые напрямую конвертируется в pandas.DataFrame.
    """
    url = request_helpers.make_url(
        engine=engine,
        market=market,
        security=security,
        ending=CANDLE_BORDERS,
    )
    table = "borders"
    return await request_helpers.get_short_data(session, url, table)


async def get_board_candle_borders(
    session: aiohttp.ClientSession,
    security: str,
    board: str = DEFAULT_BOARD,
    market: str = DEFAULT_MARKET,
    engine: str = DEFAULT_ENGINE,
) -> client.Table:
    """Получить таблицу интервалов доступных дат для указанного режиме торгов.

    Описание запроса - https://iss.moex.com/iss/reference/48

    :param session:
        Сессия http соединения.
    :param security:
        Тикер ценной бумаги
    :param board:
        Режим торгов - по умолчанию основной режим торгов T+2.
    :param market:
        Рынок - по умолчанию акции.
    :param engine:
        Движок - по умолчанию акции.

    :return:
        Список словарей, которые напрямую конвертируется в pandas.DataFrame
    """
    url = request_helpers.make_url(
        engine=engine,
        market=market,
        board=board,
        security=security,
        ending=CANDLE_BORDERS,
    )
    table = "borders"
    return await request_helpers.get_short_data(session, url, table)


async def get_market_candles(
    session: aiohttp.ClientSession,
    security: str,
    interval: int = 24,
    start: str | None = None,
    end: str | None = None,
    market: str = DEFAULT_MARKET,
    engine: str = DEFAULT_ENGINE,
) -> client.Table:
    """Получить свечи в формате HLOCV указанного инструмента на рынке для основного режима торгов.

    Если торговля идет в нескольких основных режимах, то на один интервал времени может быть выдано
    несколько свечек - по свечке на каждый режим. Предположительно такая ситуация может произойти для
    свечек длиннее 1 дня.

    Описание запроса - https://iss.moex.com/iss/reference/155

    :param session:
        Сессия http соединения.
    :param security:
        Тикер ценной бумаги.
    :param interval:
        Размер свечки - целое число 1 (1 минута), 10 (10 минут), 60 (1 час), 24 (1 день), 7 (1 неделя),
        31 (1 месяц) или 4 (1 квартал). По умолчанию дневные данные.
    :param start:
        Дата вида ГГГГ-ММ-ДД. При отсутствии данные будут загружены с начала истории.
    :param end:
        Дата вида ГГГГ-ММ-ДД. При отсутствии данные будут загружены до конца истории.
    :param market:
        Рынок - по умолчанию акции.
    :param engine:
        Движок - по умолчанию акции.

    :return:
        Список словарей, которые напрямую конвертируется в pandas.DataFrame.
    """
    url = request_helpers.make_url(engine=engine, market=market, security=security, ending=CANDLES)
    table = CANDLES
    query = request_helpers.make_query(interval=interval, start=start, end=end)
    return await request_helpers.get_long_data(session, url, table, query)


async def get_board_candles(
    session: aiohttp.ClientSession,
    security: str,
    interval: int = 24,
    start: str | None = None,
    end: str | None = None,
    board: str = DEFAULT_BOARD,
    market: str = DEFAULT_MARKET,
    engine: str = DEFAULT_ENGINE,
) -> client.Table:
    """Получить свечи в формате HLOCV указанного инструмента в указанном режиме торгов за интервал дат.

    Описание запроса - https://iss.moex.com/iss/reference/46

    :param session:
        Сессия http соединения.
    :param security:
        Тикер ценной бумаги.
    :param interval:
        Размер свечки - целое число 1 (1 минута), 10 (10 минут), 60 (1 час), 24 (1 день), 7 (1 неделя),
        31 (1 месяц) или 4 (1 квартал). По умолчанию дневные данные.
    :param start:
        Дата вида ГГГГ-ММ-ДД. При отсутствии данные будут загружены с начала истории.
    :param end:
        Дата вида ГГГГ-ММ-ДД. При отсутствии данные будут загружены до конца истории.
    :param board:
        Режим торгов - по умолчанию основной режим торгов T+2.
    :param market:
        Рынок - по умолчанию акции.
    :param engine:
        Движок - по умолчанию акции.

    :return:
        Список словарей, которые напрямую конвертируется в pandas.DataFrame.
    """
    url = request_helpers.make_url(
        engine=engine,
        market=market,
        board=board,
        security=security,
        ending=CANDLES,
    )
    table = CANDLES
    query = request_helpers.make_query(interval=interval, start=start, end=end)
    return await request_helpers.get_long_data(session, url, table, query)



import os
import re
import sys

_package_name = "aiomoex"

_route_path = os.path.abspath("../")
sys.path.insert(0, _route_path)
_version_path = os.path.abspath(os.path.join(_route_path, "pyproject.toml"))
with open(_version_path) as file:
    try:
        _version_info = re.search(
            r"^version = \"" r"(?P<major>\d+)" r"\.(?P<minor>\d+)" r"\.(?P<patch>\d+)\"$",
            file.read(),
            re.M,
        ).groupdict()
    except IndexError:
        raise RuntimeError("Unable to determine version.")

version = "{major}.{minor}".format(**_version_info)
release = "{major}.{minor}.{patch}".format(**_version_info)

extensions = ["sphinx.ext.autodoc", "sphinx.ext.githubpages"]
autodoc_member_order = "bysource"
templates_path = ["templates"]
html_static_path = ["static"]

source_suffix = ".rst"
master_doc = "index"
project = _package_name
copyright = "2018, Mikhail Korotkov aka WLMike"
author = "Mikhail Korotkov aka WLMike"
language = "ru"
exclude_patterns = ["build"]
highlight_language = "default"
html_theme = "aiohttp_theme"

html_theme_options = {
    "logo": "",
    "description": "Asyncio MOEX ISS API",
    "canonical_url": "https://wlm1ke.github.io/aiomoex",
    "github_user": "WLM1ke",
    "github_repo": "aiomoex",
    "github_button": False,
    "github_type": "",
    "github_banner": True,
    "badges": [
        {
            "image": "https://badge.fury.io/py/aiomoex.svg",
            "target": "https://badge.fury.io/py/aiomoex",
            "height": "20",
            "alt": "Latest PyPI package version",
        },
    ],
    "sidebar_collapse": False,
}

html_sidebars = {"**": ["about.html", "navigation.html"]}



"""Асинхронный клиент для MOEX ISS."""
from collections.abc import AsyncIterable, AsyncIterator
from typing import cast

import aiohttp
from aiohttp import client_exceptions

Values = str | int | float
TableRow = dict[str, Values]
Table = list[TableRow]
TablesDict = dict[str, Table]
WebQuery = dict[str, str | int]


class ISSMoexError(Exception):
    """Ошибки во время обработки запросов."""


def _cursor_block_size(start: int, cursor_table: Table) -> int:
    cursor, *wrong_data = cursor_table

    if wrong_data or cast(int, cursor["INDEX"]) != start:
        raise ISSMoexError(f"Некорректные данные history.cursor: {cursor_table}")

    block_size = cast(int, cursor["PAGESIZE"])

    if start + block_size < cast(int, cursor["TOTAL"]):
        return block_size
    return 0


class ISSClient(AsyncIterable[TablesDict]):
    """Асинхронный клиент для MOEX ISS - может быть использован с async for.

    Загружает данные для простых ответов с помощью метода get. Для ответов состоящих из нескольких блоков
    поддерживается протокол асинхронного генератора отдельных блоков или метод get_all для их
    автоматического сбора.
    """

    _client_session = None

    def __init__(self, session: aiohttp.ClientSession, url: str, query: WebQuery | None = None) -> None:
        """MOEX ISS является REST сервером.

        Полный перечень запросов и параметров к ним https://iss.moex.com/iss/reference/
        Дополнительное описание https://fs.moex.com/files/6523

        :param session:
            Сессия http соединения.
        :param url:
            Адрес запроса.
        :param query:
            Перечень дополнительных параметров запроса. К списку дополнительных параметров всегда
            добавляется требование предоставить ответ в виде расширенного json без метаданных.
        """
        self._session = session
        self._url = url
        self._query = query or {}

    def __repr__(self) -> str:
        """Наименование класса и содержание запроса к ISS Moex."""
        class_name = self.__class__.__name__
        return f"{class_name}(url={self._url}, query={self._query})"

    def __aiter__(self) -> AsyncIterator[TablesDict]:
        """Асинхронный генератор по ответам состоящим из нескольких блоков.

        На часть запросов выдается только начальный блок данных (обычно из 100 элементов). Генератор
        обеспечивает загрузку всех блоков. При этом в ответах на некоторые запросы может содержаться
        курсор с положением текущего блока данных (позволяет сэкономить один запрос). Генератор
        обеспечивает обработку ответов как с курсором, так и без него.
        """
        return self._iterator_maker()

    async def get(self, start: int | None = None) -> TablesDict:
        """Загрузка данных.

        :param start:
            Номер элемента с которого нужно загрузить данные. Используется для дозагрузки данных,
            состоящих из нескольких блоков. При отсутствии данные загружаются с начального элемента.

        :return:
            Блок данных с отброшенной вспомогательной информацией - словарь, каждый ключ которого
            соответствует одной из таблиц с данными. Таблицы являются списками словарей, которые напрямую
            конвертируются в pandas.DataFrame.
        :raises ISSMoexError:
            Ошибка при обращении к ISS Moex.
        """
        url = self._url
        query = self._make_query(start)
        async with self._session.get(url, params=query) as respond:
            try:
                respond.raise_for_status()
            except client_exceptions.ClientResponseError as err:
                raise ISSMoexError("Неверный url", respond.url) from err
            else:
                raw_respond: list[dict[str, Table]] = await respond.json()
                return raw_respond[1]

    async def get_all(self) -> TablesDict:
        """Собирает все блоки данных для запросов.

        :return:
            Объединенные из всех блоков данные с отброшенной вспомогательной информацией - словарь,
            каждый ключ которого соответствует одной из таблиц с данными. Таблицы являются списками
            словарей, которые напрямую конвертируются в pandas.DataFrame.
        """
        all_data: TablesDict = {}
        async for block in self:
            for table_name, table_rows in block.items():
                all_data.setdefault(table_name, []).extend(table_rows)
        return all_data

    def _make_query(self, start: int | None = None) -> WebQuery:
        """Формирует параметры запроса.

        К общему набору параметров запроса добавляется требование предоставить ответ в виде
        расширенного json.
        """
        query: WebQuery = {"iss.json": "extended", "iss.meta": "off"} | self._query
        if start:
            query["start"] = start
        return query

    async def _iterator_maker(self) -> AsyncIterator[TablesDict]:
        start = 0
        while True:
            respond = await self.get(start)
            if (cursor_table := respond.get("history.cursor")) is not None:
                respond.pop("history.cursor")
                yield respond

                block_size = _cursor_block_size(start, cursor_table)
            else:
                yield respond

                table_name = next(iter(respond))
                block_size = len(respond[table_name])

            if not block_size:
                return
            start += block_size



import pandas as pd
import pytest

from aiomoex import history


async def test_get_board_dates(http_session):
    data = await history.get_board_dates(http_session)
    assert isinstance(data, list)
    assert len(data) == 1
    assert data[0]["till"] >= "2023-11-24"


async def test_get_board_securities(http_session):
    data = await history.get_board_securities(http_session)
    assert isinstance(data, list)
    assert len(data) > 200
    df = pd.DataFrame(data)
    df = df.set_index("SECID")
    assert df.loc["AKRN", "SHORTNAME"] == "Акрон"
    assert df.loc["GAZP", "REGNUMBER"] == "1-02-00028-A"
    assert df.loc["TTLK", "LOTSIZE"] == 1000
    assert df.loc["MRSB", "SHORTNAME"] == "МордЭнСб"
    assert df.loc["MRSB", "REGNUMBER"] == "1-01-55055-E"
    assert df.loc["MRSB", "LOTSIZE"] == 10000


async def test_get_market_history_from_beginning(http_session):
    data = await history.get_market_history(http_session, "AKRN", end="2006-12-01")
    assert isinstance(data, list)
    assert data[0]["TRADEDATE"] == "2006-10-11"
    assert data[-1]["TRADEDATE"] == "2006-12-01"
    assert len(data[-2]) == 5
    assert "CLOSE" in data[2]
    assert "VOLUME" in data[3]


async def test_get_market_history_to_end(http_session):
    data = await history.get_market_history(http_session, "MOEX", start="2017-10-02")
    assert isinstance(data, list)
    assert len(data) > 100
    assert data[0]["TRADEDATE"] == "2017-10-02"
    assert data[-1]["TRADEDATE"] >= "2018-11-19"


async def test_get_board_history_from_beginning(http_session):
    data = await history.get_board_history(http_session, "LSNGP", end="2014-08-01")
    df = pd.DataFrame(data)
    df = df.set_index("TRADEDATE")
    assert len(df.columns) == 4
    assert df.index[0] == "2014-06-09"
    assert df.loc["2014-06-09", "CLOSE"] == pytest.approx(14.7)
    assert df.loc["2014-08-01", "VOLUME"] == 4000


async def test_get_board_history_to_end(http_session):
    data = await history.get_board_history(http_session, "LSRG", start="2018-08-07")
    df = pd.DataFrame(data)
    df = df.set_index("TRADEDATE")
    assert len(df.columns) == 4
    assert df.index[0] == "2018-08-07"
    assert df.index[-1] >= "2018-11-19"
    assert df.loc["2018-08-07", "CLOSE"] == 777
    assert df.loc["2018-08-10", "VOLUME"] == 11313
    assert df.loc["2018-08-10", "BOARDID"] == "TQBR"
    assert df.loc["2018-08-10", "VALUE"] == pytest.approx(8_626_464.5)
    assert df.loc["2018-09-06", "CLOSE"] == pytest.approx(660)
    assert df.loc["2018-08-28", "VOLUME"] == 47428



"""Реализация части запросов к MOEX ISS.

Результаты запросов являются списками словарей и могут быть конвертированы в pandas.DataFrame.

Работа функций-запросов базируется на универсальном клиенте, позволяющем осуществлять произвольные
запросы к ISS MOEX, поэтому перечень доступных функций-запросов может быть при необходимости дополнен:

- Полный перечень запросов https://iss.moex.com/iss/reference/
- Дополнительное описание https://fs.moex.com/files/6523
"""
from aiomoex.candles import (
    get_board_candle_borders,
    get_board_candles,
    get_market_candle_borders,
    get_market_candles,
)
from aiomoex.client import ISSClient, TableRow, TablesDict, Values
from aiomoex.history import get_board_dates, get_board_history, get_board_securities, get_market_history
from aiomoex.reference import find_securities, get_reference

__all__ = [
    "get_board_candle_borders",
    "get_board_candles",
    "get_market_candle_borders",
    "get_market_candles",
    "ISSClient",
    "TableRow",
    "TablesDict",
    "Values",
    "get_board_dates",
    "get_board_history",
    "get_board_securities",
    "get_market_history",
    "find_securities",
    "get_reference",
]



import pytest

from aiomoex import client, request_helpers


def test_make_query_empty():
    query = request_helpers.make_query()
    assert query == {}


def test_make_query_full():
    query = request_helpers.make_query(start=1, end=2, table=3, columns=("4",))
    assert isinstance(query, dict)
    assert len(query) == 4
    assert query["from"] == 1
    assert query["till"] == 2
    assert query["iss.only"] == "3,history.cursor"
    assert query[f"{3}.columns"] == "4"


def test_make_query_many_columns():
    query = request_helpers.make_query(table=1, columns=("2", "3"))
    assert isinstance(query, dict)
    assert len(query) == 2
    assert query["iss.only"] == "1,history.cursor"
    assert query[f"{1}.columns"] == "2,3"


def test_get_table_notable():
    with pytest.raises(client.ISSMoexError) as error:
        request_helpers.get_table({"a": "b"}, "b")
    assert "Отсутствует таблица b в данных" in str(error.value)



"""Функции для получения справочной информации."""
from collections.abc import Iterable

import aiohttp

from aiomoex import client, request_helpers
from aiomoex.request_helpers import SECURITIES


async def get_reference(session: aiohttp.ClientSession, placeholder: str = "boards") -> client.Table:
    """Получить перечень доступных значений плейсхолдера в адресе запроса.

    Например в описание запроса https://iss.moex.com/iss/reference/32 присутствует следующий адрес
    /iss/engines/[engine]/markets/[market]/boards/[board]/securities с плейсхолдерами engines, markets и
    boards.

    Описание запроса - https://iss.moex.com/iss/reference/28

    :param session:
        Сессия http соединения.
    :param placeholder:
        Наименование плейсхолдера в адресе запроса: engines, markets, boards, boardgroups, durations,
        securitytypes, securitygroups, securitycollections.

    :return:
        Список словарей, которые напрямую конвертируется в pandas.DataFrame.
    """
    url = request_helpers.make_url(ending="index")
    return await request_helpers.get_short_data(session, url, placeholder)


async def find_securities(
    session: aiohttp.ClientSession,
    string: str,
    columns: Iterable[str] | None = ("secid", "regnumber"),
) -> client.Table:
    """Найти инструменты по части Кода, Названию, ISIN, Идентификатору Эмитента, Номеру гос.регистрации.

    Один из вариантов использования - по регистрационному номеру узнать предыдущие тикеры эмитента, и с
    помощью нескольких запросов об истории котировок собрать длинную историю с использованием всех
    предыдущих тикеров.

    Описание запроса - https://iss.moex.com/iss/reference/5

    :param session:
        Сессия http соединения.
    :param string:
        Часть Кода, Названия, ISIN, Идентификатора Эмитента, Номера гос.регистрации.
    :param columns:
        Кортеж столбцов, которые нужно загрузить - по умолчанию тикер и номер государственно регистрации.
        Если пустой или None, то загружаются все столбцы.

    :return: Список словарей, которые напрямую конвертируется в pandas.DataFrame.
    """
    url = request_helpers.make_url(ending=SECURITIES)
    table = SECURITIES
    query = request_helpers.make_query(question=string, table=table, columns=columns)
    return await request_helpers.get_short_data(session, url, table, query)



import typing

import pytest

from aiomoex import client


def test_iss_client_async_iterable(http_session):
    iss = client.ISSClient(http_session, "test_url")
    assert isinstance(iss, typing.AsyncIterable)
    assert str(iss) == "ISSClient(url=test_url, query={})"


def test_make_query_empty(http_session):
    iss = client.ISSClient(http_session, "test_url")
    # noinspection PyProtectedMember
    query = iss._make_query()
    assert isinstance(query, typing.Mapping)
    assert len(query) == 2
    assert query["iss.json"] == "extended"
    assert query["iss.meta"] == "off"


def test_make_query_not_empty(http_session):
    iss = client.ISSClient(http_session, "test_url", {"test_param": "test_value"})
    # noinspection PyProtectedMember
    query = iss._make_query()
    assert isinstance(query, typing.Mapping)
    assert len(query) == 3
    assert query["iss.json"] == "extended"
    assert query["iss.meta"] == "off"
    assert query["test_param"] == "test_value"


def test_make_query_not_empty_with_start(http_session):
    iss = client.ISSClient(http_session, "test_url", {"test_param": "test_value"})
    # noinspection PyProtectedMember
    query = iss._make_query(100)
    assert isinstance(query, typing.Mapping)
    assert len(query) == 4
    assert query["iss.json"] == "extended"
    assert query["iss.meta"] == "off"
    assert query["test_param"] == "test_value"
    assert query["start"] == 100


async def test_get(http_session):
    url = "https://iss.moex.com/iss/securities.json"
    query = {"q": "1-02-65104-D"}
    iss = client.ISSClient(http_session, url, query)
    raw = await iss.get()
    data = raw["securities"]
    assert isinstance(data, list)
    assert len(data) == 4
    assert isinstance(data[0], dict)
    assert data[1]["regnumber"] == "1-02-65104-D"
    assert "Юнипро" in data[2]["emitent_title"]


async def test_get_with_start(http_session):
    url = "https://iss.moex.com/iss/securities.json"
    query = {"q": "1-02-65104-D"}
    iss = client.ISSClient(http_session, url, query)
    raw = await iss.get(1)
    data = raw["securities"]
    assert isinstance(data, list)
    assert len(data) == 3
    assert isinstance(data[0], dict)
    assert data[1]["regnumber"] == "1-02-65104-D"
    assert "Юнипро" in data[2]["emitent_title"]


async def test_get_error(http_session):
    url = "https://iss.moex.com/iss/securities1.json"
    iss = client.ISSClient(http_session, url)
    with pytest.raises(client.ISSMoexError) as error:
        await iss.get()
    assert "Неверный url" in str(error.value)
    assert "url" in str(error.value)


async def test_get_all_with_cursor(http_session):
    url = "https://iss.moex.com/iss/history/engines/stock/markets/shares/securities/SNGSP.json"
    query = {"from": "2018-01-01", "till": "2018-03-01"}
    iss = client.ISSClient(http_session, url, query)
    raw = await iss.get_all()
    data = raw["history"]
    assert isinstance(data, list)
    assert len(data) > 100
    assert data[0]["TRADEDATE"] == "2018-01-03"
    assert data[-1]["TRADEDATE"] == "2018-03-01"
    for row in data:
        for column in ["TRADEDATE", "OPEN", "LOW", "HIGH", "CLOSE", "VOLUME"]:
            assert column in row


async def test_get_all_without_cursor(http_session):
    url = "https://iss.moex.com/iss/history/engines/stock/markets/shares/boards/TQBR/securities/SNGSP.json"
    query = {"from": "2018-01-03", "till": "2018-06-01"}
    iss = client.ISSClient(http_session, url, query)
    raw = await iss.get_all()
    data = raw["history"]
    assert isinstance(data, list)
    assert len(data) > 100
    assert data[0]["TRADEDATE"] == "2018-01-03"
    assert data[-1]["TRADEDATE"] == "2018-06-01"
    for row in data:
        for column in ["TRADEDATE", "OPEN", "LOW", "HIGH", "CLOSE", "VOLUME"]:
            assert column in row


def test_cursor_block_size():
    assert client._cursor_block_size(5, [{"INDEX": 5, "PAGESIZE": 8, "TOTAL": 14}]) == 8


def test_cursor_block_size_zero():
    assert client._cursor_block_size(5, [{"INDEX": 5, "PAGESIZE": 8, "TOTAL": 13}]) == 0


def test_cursor_block_size_raise_bad_start():
    with pytest.raises(client.ISSMoexError) as error:
        client._cursor_block_size(5, [{"INDEX": 6, "PAGESIZE": 8, "TOTAL": 14}])
    assert "Некорректные данные history.cursor" in str(error.value)


def test_cursor_block_size_raise_bad_structure():
    with pytest.raises(client.ISSMoexError) as error:
        client._cursor_block_size(5, [{}, {}])
    assert "Некорректные данные history.cursor" in str(error.value)



"""Функции для получения данных об исторических дневных котировках."""
from collections.abc import Iterable

import aiohttp

from aiomoex import client, request_helpers
from aiomoex.request_helpers import DEFAULT_BOARD, DEFAULT_ENGINE, DEFAULT_MARKET, SECURITIES


async def get_board_dates(
    session: aiohttp.ClientSession,
    board: str = DEFAULT_BOARD,
    market: str = DEFAULT_MARKET,
    engine: str = DEFAULT_ENGINE,
) -> client.Table:
    """Получить интервал дат, доступных в истории для рынка по заданному режиму торгов.

    Описание запроса - https://iss.moex.com/iss/reference/26

    :param session:
        Сессия http соединения.
    :param board:
        Режим торгов - по умолчанию основной режим торгов T+2.
    :param market:
        Рынок - по умолчанию акции.
    :param engine:
        Движок - по умолчанию акции.

    :return:
        Список из одного элемента - словаря с ключами 'from' и 'till'.
    """
    url = request_helpers.make_url(
        history=True,
        engine=engine,
        market=market,
        board=board,
        ending="dates",
    )
    table = "dates"
    return await request_helpers.get_short_data(session, url, table)


async def get_board_securities(
    session: aiohttp.ClientSession,
    table: str = SECURITIES,
    columns: Iterable[str] | None = ("SECID", "REGNUMBER", "LOTSIZE", "SHORTNAME"),
    board: str = DEFAULT_BOARD,
    market: str = DEFAULT_MARKET,
    engine: str = DEFAULT_ENGINE,
) -> client.Table:
    """Получить таблицу инструментов по режиму торгов со вспомогательной информацией.

    Описание запроса - https://iss.moex.com/iss/reference/32

    :param session:
        Сессия http соединения.
    :param table:
        Таблица с данными, которую нужно вернуть: securities - справочник торгуемых ценных бумаг,
        marketdata - данные с результатами торгов текущего дня.
    :param columns:
        Кортеж столбцов, которые нужно загрузить - по умолчанию тикер, номер государственно регистрации,
        размер лота и краткое название. Если пустой или None, то загружаются все столбцы.
    :param board:
        Режим торгов - по умолчанию основной режим торгов T+2.
    :param market:
        Рынок - по умолчанию акции.
    :param engine:
        Движок - по умолчанию акции.

    :return:
        Список словарей, которые напрямую конвертируется в pandas.DataFrame
    """
    url = request_helpers.make_url(engine=engine, market=market, board=board, ending=SECURITIES)
    query = request_helpers.make_query(table=table, columns=columns)
    return await request_helpers.get_short_data(session, url, table, query)


async def get_market_history(
    session: aiohttp.ClientSession,
    security: str,
    start: str | None = None,
    end: str | None = None,
    columns: Iterable[str] | None = ("BOARDID", "TRADEDATE", "CLOSE", "VOLUME", "VALUE"),
    market: str = DEFAULT_MARKET,
    engine: str = DEFAULT_ENGINE,
) -> client.Table:
    """Получить историю по одной бумаге на рынке для всех режимов торгов за интервал дат.

    На одну дату может приходиться несколько значений, если торги шли в нескольких режимах.

    Описание запроса - https://iss.moex.com/iss/reference/63

    :param session:
        Сессия http соединения.
    :param security:
        Тикер ценной бумаги.
    :param start:
        Дата вида ГГГГ-ММ-ДД. При отсутствии данные будут загружены с начала истории.
    :param end:
        Дата вида ГГГГ-ММ-ДД. При отсутствии данные будут загружены до конца истории.
    :param columns:
        Кортеж столбцов, которые нужно загрузить - по умолчанию режим торгов, дата торгов, цена закрытия
        и объем в штуках и стоимости. Если пустой или None, то загружаются все столбцы.
    :param market:
        Рынок - по умолчанию акции.
    :param engine:
        Движок - по умолчанию акции.

    :return:
        Список словарей, которые напрямую конвертируется в pandas.DataFrame.
    """
    url = request_helpers.make_url(history=True, engine=engine, market=market, security=security)
    table = "history"
    query = request_helpers.make_query(start=start, end=end, table=table, columns=columns)
    return await request_helpers.get_long_data(session, url, table, query)


async def get_board_history(
    session: aiohttp.ClientSession,
    security: str,
    start: str | None = None,
    end: str | None = None,
    columns: Iterable[str] | None = ("BOARDID", "TRADEDATE", "CLOSE", "VOLUME", "VALUE"),
    board: str = DEFAULT_BOARD,
    market: str = DEFAULT_MARKET,
    engine: str = DEFAULT_ENGINE,
) -> client.Table:
    """Получить историю торгов для указанной бумаги в указанном режиме торгов за указанный интервал дат.

    Описание запроса - https://iss.moex.com/iss/reference/65

    :param session:
        Сессия http соединения.
    :param security:
        Тикер ценной бумаги.
    :param start:
        Дата вида ГГГГ-ММ-ДД. При отсутствии данные будут загружены с начала истории.
    :param end:
        Дата вида ГГГГ-ММ-ДД. При отсутствии данные будут загружены до конца истории.
    :param columns:
        Кортеж столбцов, которые нужно загрузить - по умолчанию режим торгов, дата торгов, цена закрытия
        и объем в штуках и стоимости. Если пустой или None, то загружаются все столбцы.
    :param board:
        Режим торгов - по умолчанию основной режим торгов T+2.
    :param market:
        Рынок - по умолчанию акции.
    :param engine:
        Движок - по умолчанию акции.

    :return:
        Список словарей, которые напрямую конвертируется в pandas.DataFrame.
    """
    url = request_helpers.make_url(
        history=True,
        engine=engine,
        market=market,
        board=board,
        security=security,
    )
    table = "history"
    query = request_helpers.make_query(start=start, end=end, table=table, columns=columns)
    return await request_helpers.get_long_data(session, url, table, query)



import pytest

from aiomoex import candles


async def test_get_market_candle_borders(http_session):
    data = await candles.get_market_candle_borders(http_session, "SNGSP")
    assert isinstance(data, list)
    assert len(data) == 7
    for i in data:
        del i["end"]
    assert data == [
        {"begin": "2011-12-15 10:00:00", "interval": 1, "board_group_id": 57},
        {"begin": "2003-07-01 00:00:00", "interval": 4, "board_group_id": 57},
        {"begin": "2003-07-28 00:00:00", "interval": 7, "board_group_id": 57},
        {"begin": "2011-12-08 10:00:00", "interval": 10, "board_group_id": 57},
        {"begin": "2003-07-31 00:00:00", "interval": 24, "board_group_id": 57},
        {"begin": "2003-07-01 00:00:00", "interval": 31, "board_group_id": 57},
        {"begin": "2011-11-17 10:00:00", "interval": 60, "board_group_id": 57},
    ]


async def test_get_board_candle_borders(http_session):
    data = await candles.get_board_candle_borders(http_session, "UPRO")
    assert isinstance(data, list)
    assert len(data) == 7
    for i in data:
        del i["end"]
    assert data == [
        {"begin": "2016-07-01 09:59:00", "interval": 1},
        {"begin": "2016-07-01 00:00:00", "interval": 4},
        {"begin": "2016-06-27 00:00:00", "interval": 7},
        {"begin": "2016-07-01 09:50:00", "interval": 10},
        {"begin": "2016-07-01 00:00:00", "interval": 24},
        {"begin": "2016-07-01 00:00:00", "interval": 31},
        {"begin": "2016-07-01 09:00:00", "interval": 60},
    ]


async def test_get_market_candles_from_beginning(http_session):
    data = await candles.get_market_candles(http_session, "RTKM", interval=1, end="2011-12-16")
    assert isinstance(data, list)
    assert len(data) > 500
    assert len(data[0]) == 8
    assert data[0]["open"] == pytest.approx(141.55)
    assert data[1]["close"] == pytest.approx(141.59)
    assert data[2]["high"] == pytest.approx(142.4)
    assert data[3]["low"] == pytest.approx(140.81)
    assert data[4]["value"] == pytest.approx(2_586_296.9)
    assert data[5]["volume"] == pytest.approx(4140)
    assert data[6]["begin"] == "2011-12-15 10:06:00"
    assert data[-1]["end"] == "2011-12-16 18:44:59"


async def test_get_market_candles_to_end(http_session):
    data = await candles.get_market_candles(http_session, "LSRG", interval=24, start="2020-08-20")
    assert isinstance(data, list)
    assert len(data) > 13
    assert len(data[0]) == 8
    assert data[0]["open"] == pytest.approx(775.4)
    assert data[1]["close"] == pytest.approx(771.8)
    assert data[2]["high"] == pytest.approx(779.8)
    assert data[3]["low"] == pytest.approx(770.2)
    assert data[4]["value"] == pytest.approx(59495740.6)
    assert data[6]["begin"] == "2020-08-28 00:00:00"


async def test_get_market_candles_empty(http_session):
    data = await candles.get_market_candles(http_session, "KSGR", interval=24)
    assert isinstance(data, list)
    assert len(data) == 0


async def test_get_board_candles_from_beginning(http_session):
    data = await candles.get_board_candles(http_session, "MTSS", interval=10, end="2011-12-22")
    assert isinstance(data, list)
    assert len(data) > 500
    assert len(data[0]) == 8
    assert data[0]["open"] == pytest.approx(202.7)
    assert data[1]["close"] == pytest.approx(204.12)
    assert data[2]["high"] == pytest.approx(205)
    assert data[3]["low"] == pytest.approx(204.93)
    assert data[4]["value"] == pytest.approx(3_990_683.9)
    assert data[5]["volume"] == pytest.approx(3000)
    assert data[6]["begin"] == "2011-12-08 11:00:00"
    assert data[-1]["end"] == "2011-12-22 18:49:59"


async def test_get_board_candles_to_end(http_session):
    data = await candles.get_board_candles(http_session, "TTLK", interval=31, start="2014-07-01")
    assert isinstance(data, list)
    assert len(data) > 52
    assert len(data[0]) == 8
    assert data[0]["open"] == pytest.approx(0.152)
    assert data[1]["close"] == pytest.approx(0.15689)
    assert data[2]["high"] == pytest.approx(0.15998)
    assert data[3]["low"] == pytest.approx(0.149)
    assert data[4]["value"] == pytest.approx(2_713_625)
    assert data[5]["volume"] == pytest.approx(20_180_000)
    assert data[6]["begin"] == "2015-01-01 00:00:00"
    assert data[51]["end"] == "2018-10-31 00:00:00"



import pytest

from aiomoex import reference


async def test_get_reference(http_session):
    data = await reference.get_reference(http_session, "engines")
    assert isinstance(data, list)
    assert len(data) == 10
    assert data[0] == {"id": 1, "name": "stock", "title": "Фондовый рынок и рынок депозитов"}


check_points = [
    ("1-02-65104-D", {"UPRO", "EONR", "OGK4"}),
    ("10301481B", {"SBER", "SBER03"}),
    ("20301481B", {"SBERP", "SBERP03"}),
    ("1-02-06556-A", {"PHOR"}),
]


@pytest.mark.parametrize(("reg_number", "expected"), check_points)
async def test_find_find_securities(http_session, reg_number, expected):
    data = await reference.find_securities(http_session, reg_number)
    assert isinstance(data, list)
    assert expected == {row["secid"] for row in data if row["regnumber"] == reg_number}



"""Вспомогательные функции для построения запросов."""
from collections.abc import Iterable
from typing import Final

import aiohttp

from aiomoex import client

# Режимы по умолчанию для запросов
DEFAULT_ENGINE: Final = "stock"
DEFAULT_MARKET: Final = "shares"
DEFAULT_BOARD: Final = "TQBR"
# Ключевые плейсхолдеры и константы для запросов
SECURITIES: Final = "securities"
CANDLE_BORDERS: Final = "candleborders"
CANDLES: Final = "candles"


def make_url(
    *,
    history: bool | None = None,
    engine: str | None = None,
    market: str | None = None,
    board: str | None = None,
    security: str | None = None,
    ending: str | None = None,
) -> str:
    """Формирует URL для запроса."""
    url_parts = ["https://iss.moex.com/iss"]
    if history:
        url_parts.append("/history")
    if engine:
        url_parts.append(f"/engines/{engine}")
    if market:
        url_parts.append(f"/markets/{market}")
    if board:
        url_parts.append(f"/boards/{board}")
    if security:
        url_parts.append(f"/securities/{security}")
    if ending:
        url_parts.append(f"/{ending}")
    url_parts.append(".json")
    return "".join(url_parts)


def make_query(
    *,
    question: str | None = None,
    interval: int | None = None,
    start: str | None = None,
    end: str | None = None,
    table: str | None = None,
    columns: Iterable[str] | None = None,
) -> client.WebQuery:
    """Формирует дополнительные параметры запроса к MOEX ISS.

    :param question:
        Строка с частью характеристик бумаги для поиска.
    :param interval:
        Размер свечки.
    :param start:
        Начальная дата котировок.
    :param end:
        Конечная дата котировок.
    :param table:
        Таблица, которую нужно загрузить (для запросов, предполагающих наличие нескольких таблиц).
    :param columns:
        Кортеж столбцов, которые нужно загрузить.

    :return:
        Словарь с дополнительными параметрами запроса.
    """
    query: client.WebQuery = {}
    if question:
        query["q"] = question
    if interval:
        query["interval"] = interval
    if start:
        query["from"] = start
    if end:
        query["till"] = end
    if table:
        query["iss.only"] = f"{table},history.cursor"
    if columns:
        query[f"{table}.columns"] = ",".join(columns)
    return query


def get_table(table_dict: client.TablesDict, table_name: str) -> client.Table:
    """Извлекает конкретную таблицу из данных."""
    try:
        table = table_dict[table_name]
    except KeyError as err:
        raise client.ISSMoexError(f"Отсутствует таблица {table_name} в данных") from err
    return table


async def get_short_data(
    session: aiohttp.ClientSession,
    url: str,
    table_name: str,
    query: client.WebQuery | None = None,
) -> client.Table:
    """Получить данные для запроса с выдачей всей информации за раз.

    :param session:
        Сессия http соединения.
    :param url:
        URL запроса.
    :param query:
        Дополнительные параметры запроса - None, если нет параметров.
    :param table_name:
        Таблица, которую нужно выбрать.

    :return:
        Конкретная таблица из запроса.
    """
    iss = client.ISSClient(session, url, query)
    table_dict = await iss.get()
    return get_table(table_dict, table_name)


async def get_long_data(
    session: aiohttp.ClientSession,
    url: str,
    table_name: str,
    query: client.WebQuery | None = None,
) -> client.Table:
    """Получить данные для запроса, в котором информация выдается несколькими блоками.

    :param session:
        Сессия http соединения.
    :param url:
        URL запроса.
    :param query:
        Дополнительные параметры запроса - None, если нет параметров.
    :param table_name:
        Таблица, которую нужно выбрать.

    :return:
        Конкретная таблица из запроса.
    """
    iss = client.ISSClient(session, url, query)
    table_dict = await iss.get_all()
    return get_table(table_dict, table_name)
